@@ -1,14 +1,15 @@
 //! This pass removes jumps to basic blocks containing only a return, and replaces them with a
 //! return instead.
 
-use crate::simplify;
 use rustc_index::bit_set::BitSet;
 use rustc_middle::mir::*;
 use rustc_middle::ty::TyCtxt;
 
-pub struct MultipleReturnTerminators;
+use crate::simplify;
 
-impl<'tcx> MirPass<'tcx> for MultipleReturnTerminators {
+pub(super) struct MultipleReturnTerminators;
+
+impl<'tcx> crate::MirPass<'tcx> for MultipleReturnTerminators {
     fn is_enabled(&self, sess: &rustc_session::Session) -> bool {
         sess.mir_opt_level() >= 4
     }
